from .core import Graph
from typing import List


def annotation(namespace: str, name: str = '') -> Graph:
    """Retrieve a list of BEL statements defined by a given namespace and name/term.

    :param str namespace: The namespace of the given name/term/value e.g. 'HGNC' or 'MGI'.
    :param str name: The term or value e.g. a protein symbol or MeSH term.
    :return: Graph of the results
    :rtype: Graph
    """
    return Graph().apply_api_function('_bel_by_annotation', namespace, name)


def authors(author_list: List[str], edge_class: str = '', node_class: str = '', exclude_namespace: str = '') -> Graph:
    """Retrieve a list of BEL statements defined by a list of authors. For greater specificity, an edge and node class
    can be passed to filter results as well as namespaces to exclude.

    :param List[str] author_list: List of author names of publications.
    :param str edge_class: Desired edge class to filter results by.
    :param str node_class: Desired node class to filter results by.
    :param str exclude_namespace: Namespace to exclude in result set.
    :return: Graph of the results.
    :rtype: Graph
    """
    authors_query_str = ",".join(author_list)
    return Graph().apply_api_function('_bel_by_authors',
                                      authors_query_str,
                                      edge_class,
                                      node_class,
                                      exclude_namespace)


def last_author(namespace: str, name: str = '') -> Graph:
    return Graph().apply_api_function('_bel_by_last_author', namespace, name)


def pmid(pmid: int) -> Graph:
    return Graph().apply_api_function('_bel_by_pmid', pmid)


def subgraph(namespace: str, name: str = '') -> Graph:
    return Graph().apply_api_function('_bel_by_subgraph', namespace, name)


def causal_correlative_by_gene(gene_symbol: str) -> Graph:
    return Graph().apply_api_function('_bel_causal_correlative_by_gene', gene_symbol)


def path(from_gene: str, to_gene: str, min_edges: int = 1, max_edges: int = 4) -> Graph:
    num_range = f"{min_edges}-{max_edges}"
    return Graph().apply_api_function('_bel_path', from_gene, to_gene, num_range)


def belish(statement: str) -> Graph:
    return Graph().apply_api_function('_belish', statement)


# TODO: Implemnet this in ebel
# def get_all_causal():
#     return Graph().apply_api_function('get_all_causal')



"""Testing module for query"""
from ebel_rest import query, connect
from ..constants import USER, PASSWORD, DATABASE, SERVER


class TestQuery:

    def test_annotation(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        q = query.annotation('MeSHAnatomy', name='Lung')
        assert len(q.edges) > 1

    def test_last_author(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        q = query.last_author(['Hong W'])
        assert len(q.table.columns) >= 4
        assert len(q.table.index) > 0

    def test_pmid(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        q = query.pmid(30310104)
        assert len(q.table.columns) >= 4
        assert len(q.table.index) > 0

    def test_causal_correlative_by_gene(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        q = query.causal_correlative_by_gene('IL6')
        assert len(q.table.columns) >= 4
        assert len(q.table.index) > 0

    def test_path(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        q = query.path("ACE2", "COVID-19", 1, 2)
        assert len(q.table.columns) >= 4
        assert len(q.table.index) > 0

    def test_belish(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        q = query.belish('p(HGNC:"ACE2") ? ?')
        assert len(q.table.columns) >= 4
        assert len(q.table.index) > 0

# TODO write tests for "subgraph"



"""Main module."""
import os
import re
import json

import graphviz
import urllib.parse
import urllib.request
import pandas as pd
from IPython.display import display, Image

from .visualisation.colours.graphviz import edge_colours, node_colours
from .defaults import pics_path


pd.set_option('display.max_colwidth', None)


class Connector:
    user = None
    password = None
    server = None
    db_name = None
    print_url = False


def connect(user, password, server, db_name, print_url=False) -> None:
    """Connects to the database.

    :param str user: Database username.
    :param str password: Database password for specified username.
    :param str server: Server or URL where database is hosted.
    :param str db_name: Name of database.
    :param bool print_url: Boolean to choose whether to print the REST API call.
    """
    Connector.user = user
    Connector.password = password
    Connector.server = server
    Connector.db_name = db_name
    Connector.print_url = print_url


class Client:
    def __init__(self):
        self._user = Connector.user
        self.__passwd = Connector.password
        self.url_template = f"{Connector.server}/function/{Connector.db_name}/{{function_name}}/{{arguments}}"
        self._data = None
        self.function_name = None
        self.print_url = Connector.print_url

    def _get_data(self, function_name, *args):
        parameters = []
        for arg in args:
            parameters.append(urllib.parse.quote(str(arg)))
        url = self.url_template.format(
            function_name=function_name,
            arguments='/'.join(parameters))

        passman = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        passman.add_password(None, url, self._user, self.__passwd)
        authhandler = urllib.request.HTTPBasicAuthHandler(passman)
        opener = urllib.request.build_opener(authhandler)
        urllib.request.install_opener(opener)
        if self.print_url:
            print(url)
        res = urllib.request.urlopen(url)
        res_body = res.read()

        self._data = json.loads(res_body)['result']

    def apply_api_function(self, function_name, *args):
        self.function_name = function_name
        self._get_data(function_name, *args)
        return self

    @property
    def data(self):
        return [{k: v for k, v in x.items() if not k.startswith('@')} for x in self._data]

    @property
    def table(self):
        """Returns pandas dataframe."""
        if len(self._data):
            if 'edge_id' in self._data[0].keys():
                cols = ['subject_bel', 'relation', 'object_bel', 'pmid', 'edge_id']
                results = [tuple([x[col] for col in cols]) for x in self._data]
                df = pd.DataFrame(results, columns=cols)
                df.set_index('edge_id', inplace=True)
            else:
                results = [{k: v for k, v in x.items() if not k.startswith('@')} for x in self._data]
                df = pd.DataFrame(results)
            return df
        return "No results"


class Statistics(Client):
    # in the moment only a copy of Client
    pass


class Graph(Client):
    @property
    def edge_ids(self):
        return set([x['edge_id'] for x in self._data])

    @property
    def edges(self):
        data_unique = {x['edge_id']: x for x in self._data}
        return list(data_unique.values())

    def __xor__(self, other):
        """
        Return new graph with edges in either this or the other Graph object but not both.
        :param other: BEL graph
        :return: BEL graph
        """
        if isinstance(other, Graph):
            new_graph = Graph()
            set_edge_ids = self.edge_ids.symmetric_difference(other.edge_ids)
            union_edges = self.edges + other.edges
            new_graph._data = [x for x in union_edges if x['edge_id'] in set_edge_ids]
            new_graph.function_name = "joined_graph"
            return new_graph
        else:
            raise Exception('Second element is not a graph')

    def __or__(self, other):
        """
        Return new graph with edges in both graphs.
        :param other: BEL graph
        :return: BEL graph
        """
        return self.__add__(other)

    def __add__(self, other):
        """
        Return new graph with edges in both graphs.
        :param other: BEL graph
        :return: BEL graph
        """
        if isinstance(other, Graph):
            new_graph = Graph()
            set_edge_ids = self.edge_ids.union(other.edge_ids)
            union_edges = self.edges + other.edges
            new_graph._data = [x for x in union_edges if x['edge_id'] in set_edge_ids]
            new_graph.function_name = "joined_graph"
            return new_graph
        else:
            raise Exception('Second element is not a graph')

    def __sub__(self, other):
        """
        Return new graph with edges in this, but not the other graph.
        :param other: BEL graph
        :return: BEL graph
        """
        if isinstance(other, Graph):
            new_graph = Graph()
            set_edge_ids = self.edge_ids.difference(other.edge_ids)
            union_edges = self.edges + other.edges
            new_graph._data = [x for x in union_edges if x['edge_id'] in set_edge_ids]
            new_graph.function_name = "subtracted_graph"
            return new_graph
        else:
            raise Exception('Second element is not a graph')

    def __and__(self, other):
        """
        Return new graph with edges common to both graphs.
        :param other: BEL graph
        :return: BEL graph
        """
        if isinstance(other, Graph):
            new_graph = Graph()
            set_edge_ids = self.edge_ids.intersection(other.edge_ids)
            union_edges = self.edges + other.edges
            new_graph._data = [x for x in union_edges if x['edge_id'] in set_edge_ids]
            new_graph.function_name = "unioned_graph"
            return new_graph
        else:
            raise Exception('Second element is not a graph')

    def __ge__(self, other) -> bool:
        """Test whether every this graph is a supergraph of other graph."""
        if isinstance(other, Graph):
            return self.edge_ids.issuperset(other.edge_ids)
        else:
            raise Exception('Second element is not a graph')

    def __le__(self, other) -> bool:
        """Test whether every edge in this graph is in other graph."""
        if isinstance(other, Graph):
            return self.edge_ids.issubset(other.edge_ids)
        else:
            raise Exception('Second element is not a graph')

    def __len__(self) -> int:
        """
        Return number of edges.
        :return: int
        """
        return len(self.edge_ids)

    def __eq__(self, other):
        """
        Return true if both graphs are equivalent.
        :param other: BEL graph
        :return:
        """
        if isinstance(other, Graph):
            return self.edge_ids == other.edge_ids
        else:
            raise Exception('Second element is not a graph')

    def __ne__(self, other):
        """
        Return true if both graphs are not equivalent.
        :param other: BEL graph
        :return:
        """
        if isinstance(other, Graph):
            return self.edge_ids != other.edge_ids
        else:
            raise Exception('Second element is not a graph')

    def as_graph(self):
        self._ebel_graph(False, False)

    def as_graph_with_ids(self):
        self._ebel_graph(True, False)

    def as_graph_bel(self):
        self._ebel_graph(False, True)

    def as_graph_bel_with_ids(self):
        self._ebel_graph(True, True)

    def _ebel_graph(self, with_edge_id, bel_names):
        rs = self.edges
        d = graphviz.Digraph(format='png')
        d.attr(size="300,300")
        d.attr('node', shape='box')
        d.attr('node', style='filled')

        node_names = []

        for i in range(len(rs)):
            for so in ['subject', 'object']:
                node_id = rs[i][f'{so}_id'] = rs[i][f'{so}_id'].replace(':', '.')
                if bel_names:
                    node_label = re.sub('"', '\"', rs[i][so + "_bel"])
                else:
                    involved = rs[i][so + "_involved_genes"] + rs[i][so + "_involved_other"]
                    node_label = ', '.join(involved)
                    node_label = rs[i][so + "_class"] + "\n" + node_label
                node_class = rs[i][so + "_class"]
                d.attr('node', fillcolor=node_colours.get(node_class, 'grey'))
                d.node(node_id, node_label)
                node_names.append(node_id)

        for r in rs:
            d.attr('edge', color=edge_colours.get(r['relation'], 'grey'))
            if with_edge_id:
                edge_lable = f"{r['relation']} {r['edge_id']}"
            else:
                edge_lable = r['relation']
            d.edge(r['subject_id'], r['object_id'], edge_lable)

        file_path = d.render(os.path.join(pics_path, self.function_name))
        display(Image(filename=file_path))

    @property
    def table_all_columns(self):
        """Returns pandas dataframe."""
        cols = ['subject_bel',
                'relation',
                'object_bel',
                'annotation',
                'last_author',
                'publication_date',
                'title',
                'evidence',
                'pmid',
                'edge_id']
        if len(self._data):
            results = [tuple([x[col] for col in cols]) for x in self._data]
            df = pd.DataFrame(results, columns=cols)
            df.set_index('edge_id', inplace=True)
            return df
        return "No results"



edge_colours = {
    'increases': "limegreen",
    'directly_increases': "springgreen4",
    'decreases': "orangered",
    'directly_decreases': "red3",
    'rate_limiting_step_of': "lightslateblue",
    'regulates': "lightblue3",
    'causes_no_change': "yellow2",
    'positive_correlation': "blue"
}

node_colours = {
    'protein': "lightblue1",
    'complex': "khaki1",
    'component': "aquamarine",
    'rna': "goldenrod1",
    'gene': "lightslateblue",
    'activity': "palegreen3",
    'abundance': 'darksalmon',
    'pathology': "palegreen",
}



"""Testing module for statistics"""
from ebel_rest import statistics, connect
from ..constants import USER, PASSWORD, DATABASE, SERVER


class TestStatistics:

    def test_summarize(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.summarize()
        assert stats.data is not None
        assert stats.table.shape == (5, 2)

    def test_publication_by_year(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.publication_by_year()
        assert stats.data is not None
        assert len(stats.table.columns) == 2
        assert len(stats.table.index) > 0

    def test_publication_by_number_of_statements(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.publication_by_number_of_statements()
        assert stats.data is not None
        assert len(stats.table.columns) == 6
        assert len(stats.table.index) > 0

    def test_last_author_by_number_of_publications(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.last_author_by_number_of_publications()
        assert stats.data is not None
        assert len(stats.table.columns) == 2
        assert len(stats.table.index) > 0

    def test_last_author_by_number_of_statements(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.last_author_by_number_of_statements()
        assert stats.data is not None
        assert len(stats.table.columns) == 2
        assert len(stats.table.index) > 0

    def test_node_namespace_order_by_count(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.node_namespace_order_by_count()
        assert stats.data is not None
        assert len(stats.table.columns) == 3
        assert len(stats.table.index) > 0

    def test_edges(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.edges()
        assert stats.data is not None
        assert len(stats.table.columns) == 2
        assert len(stats.table.index) > 0

    def test_nodes(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.nodes()
        assert stats.data is not None
        assert len(stats.table.columns) == 2
        assert len(stats.table.index) > 0

    def test_total_bel_nodes(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.total_bel_nodes()
        assert stats.data is not None
        assert stats.table.shape == (1, 1)

    def test_total_bel_edges(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.total_bel_edges()
        assert stats.data is not None
        assert stats.table.shape == (1, 1)

    def test_total_publications(self):
        connect(USER, PASSWORD, SERVER, DATABASE)
        stats = statistics.total_publications()
        assert stats.data is not None
        assert stats.table.shape == (1, 1)



"""Constant values for testing ebel_rest. Here we use the publicly available COVID-19 Knowledge
Graph login information."""

USER = "covid_user"
SERVER = "http://graphstore.scai.fraunhofer.de"
PASSWORD = "covid"
DATABASE = "covid"



from .core import Statistics


def summarize() -> Statistics:
    """Returns summary statistics on the graph."""
    return Statistics().apply_api_function('bel_statistics_summarize')


def publication_by_year() -> Statistics:
    """Returns statistics on the number of publications per year in the knowledge graph."""
    return Statistics().apply_api_function('_bel_statistics_publication_by_year')


def publication_by_number_of_statements() -> Statistics:
    """Returns statistics on the number of statements per publication in the knowledge graph."""
    return Statistics().apply_api_function('_bel_statistics_publication_by_number_of_statements')


def last_author_by_number_of_publications() -> Statistics:
    """Returns statistics on the number of publications per author in the knowledge graph."""
    return Statistics().apply_api_function('_bel_statistics_last_author_by_number_of_publications')


def last_author_by_number_of_statements() -> Statistics():
    """Returns statistics on the number of statements per author in the knowledge graph."""
    return Statistics().apply_api_function('_bel_statistics_last_author_by_number_of_statements')


def node_namespace_order_by_count() -> Statistics():
    """Returns statistics on the frequency of each node type and each namespace in the knowledge graph
    in order of count."""
    return Statistics().apply_api_function('_bel_statistics_node_namespace_order_by_count')


def node_namespace_order_by_namespace() -> Statistics():
    return Statistics().apply_api_function('_bel_statistics_node_namespace_order_by_namespace')


def edges() -> Statistics():
    """Returns statistics on the frequency of each edge type in the knowledge graph."""
    return Statistics().apply_api_function('_bel_statistics_edges')


def nodes() -> Statistics():
    """Returns statistics on the frequency of each node type in the knowledge graph."""
    return Statistics().apply_api_function('_bel_statistics_nodes')


def total_bel_nodes() -> Statistics():
    """Returns the total number of nodes generated from curated statements in the knowledge graph."""
    return Statistics().apply_api_function('bel_statistics_total_bel_nodes')


def total_bel_edges() -> Statistics():
    """Returns the total number of BEL curated edges in the knowledge graph."""
    return Statistics().apply_api_function('bel_statistics_total_stmts')


def total_publications() -> Statistics():
    """Returns the total number of publications in the knowledge graph."""
    return Statistics().apply_api_function('bel_statistics_total_publications')


def subgraphs() -> Statistics():
    return Statistics().apply_api_function('_bel_statistics_subgraph')
